# Copyright 2015 eNovance
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc

import mock
from oslo_db.sqlalchemy import test_migrations
import six
import sqlalchemy

from gnocchi.indexer import sqlalchemy_base
from gnocchi.tests import base


class ABCSkip(base.SkipNotImplementedMeta, abc.ABCMeta):
    pass


class ModelsMigrationsSync(
        six.with_metaclass(ABCSkip,
                           base.TestCase,
                           test_migrations.ModelsMigrationsSync)):

    def setUp(self):
        super(ModelsMigrationsSync, self).setUp()
        self.db = mock.Mock()

    @staticmethod
    def get_metadata():
        return sqlalchemy_base.Base.metadata

    def get_engine(self):
        return self.index.get_engine()

    @staticmethod
    def db_sync(engine):
        # NOTE(jd) Nothing to do here as setUp() in the base class is already
        # creating table using upgrade
        pass

    @staticmethod
    def filter_metadata_diff(diff):
        new_diff = []
        for line in diff:
            item = line[1]
            # NOTE(sileht): skip resource types created dynamically
            if (isinstance(item, sqlalchemy.Table)
                    and item.name.startswith("rt_")):
                continue
            elif (isinstance(item, sqlalchemy.Index)
                    and item.name.startswith("ix_rt_")):
                continue
            new_diff.append(line)
        return new_diff
